/*
 * Copyright (c) 2007 Michal Kebrt
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * - The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/** @addtogroup libcarm32	
 * @{
 */
/** @file 
 *  @brief Fibrils related declarations.
 */

#ifndef LIBC_arm32_FIBRIL_H_
#define LIBC_arm32_FIBRIL_H_

#include <sys/types.h>
#include <align.h>
#include <thread.h>

/** Size of a stack item */
#define STACK_ITEM_SIZE		4

/** Stack alignment - see <a href="http://www.arm.com/support/faqdev/14269.html">ABI</a> for details */
#define STACK_ALIGNMENT		8

#define SP_DELTA	(0 + ALIGN_UP(STACK_ITEM_SIZE, STACK_ALIGNMENT))


/** Sets data to the context. 
 *  
 *  @param c     Context (#context_t).
 *  @param _pc   Program counter.
 *  @param stack Stack address.
 *  @param size  Stack size.
 *  @param ptls  Pointer to the TCB.
 */
#define context_set(c, _pc, stack, size, ptls) \
	do { \
		(c)->pc = (sysarg_t) (_pc); \
		(c)->sp = ((sysarg_t) (stack)) + (size) - SP_DELTA; \
 		(c)->tls = ((sysarg_t)(ptls)) + sizeof(tcb_t) + ARM_TP_OFFSET; \
		(c)->fp = 0; \
	} while (0)

/** Fibril context. 
 *
 *  Only registers preserved accross function calls are included. r9 is used 
 *  to store a TLS address. -ffixed-r9 gcc forces gcc not to use this
 *  register. -mtp=soft forces gcc to use #__aeabi_read_tp to obtain
 *  TLS address.
 */
typedef struct  {
	uint32_t sp;
	uint32_t pc;
	uint32_t r4;
	uint32_t r5;
	uint32_t r6;
	uint32_t r7;
	uint32_t r8;
	uint32_t tls;	/* r9 */
	uint32_t r10;
	uint32_t fp;	/* r11 */
} context_t;

static inline uintptr_t context_get_fp(context_t *ctx)
{
	return ctx->fp;
}


#endif

/** @}
 */
